/* uncompress_db.c */

#include <Pilot.h>
#include <DataPrv.h>

#define __PALMOS__
#include "zlib.h"

#include "pz.h"


/*
 * constants
 */

#define BUFFER_SIZE 1024

#define INFLATE_HEADER_SIZE 100
#define INFLATE_CHUNK_SIZE 1024


/*
 * types
 */

struct in_stream {
    struct stream stream;

    z_stream z;
    VoidPtr buffer;

    ULong uncompressed;

    progress_callback_t progress_callback;
};


/*
 * in_stream functions
 */

static int open_stream(struct in_stream *in,
	DmOpenRef db, struct stream_info *stream, Boolean allocate_buffer,
	progress_callback_t callback)
{
    int err;

    if (allocate_buffer == false) {
	in->buffer = NULL;
    } else {
	in->buffer = MemPtrNew(BUFFER_SIZE);
	if (in->buffer == NULL) {
	    return MEMORY_ERROR;
	}
    }

    in->z.zalloc = NULL;
    in->z.zfree = NULL;
    in->z.opaque = NULL;
    err = inflateInit(&in->z);
    if (err != Z_OK) {
	if (in->buffer != NULL) {
	    MemPtrFree(in->buffer);
	}
	return ZLIB_ERROR;
    }

    in->z.next_in = open_in_stream(&in->stream, db, stream);
    in->uncompressed = stream->uncompressed;
    in->progress_callback = callback;

    return noErr;
}

static int close_stream(struct in_stream *in)
{
    int err;

    err = inflateEnd(&in->z);

    if (in->buffer != NULL) {
	MemPtrFree(in->buffer);
    }
    close_in_stream(&in->stream);

    return (err == Z_OK) ? noErr : ZLIB_ERROR;
}


/*
 * uncompressing
 */

static int inflate_block(
	struct in_stream *in, void *ptr, ULong size, ULong chunk_size)
{
    int err;

    in->z.next_out = ptr;
    in->z.avail_out = size;

    do {
	in->z.avail_in = avail_in_stream(
		&in->stream, in->z.total_in, &in->z.next_in);
	if (in->z.avail_in == 0) {
	    return DATA_ERROR;
	}
	if (in->z.avail_in > chunk_size) {
	    in->z.avail_in = chunk_size;
	}

	err = inflate(&in->z, Z_SYNC_FLUSH);
	if (err == Z_STREAM_END && in->z.avail_out == 0) {
	    break;
	} else if (err != Z_OK) {
	    return ZLIB_ERROR;
	}

	if (in->progress_callback != NULL
	 && (*in->progress_callback)(in->z.total_out, in->uncompressed) != 0) {
	    return CANCELED;
	}
    } while (in->z.avail_out > 0);

    return noErr;
}

static int inflate_handle(struct in_stream *in, VoidHand h, ULong size)
{
    VoidPtr ptr;
    ULong read_size, chunk_size;
    int err;

    ptr = MemHandleLock(h);

    read_size = 0;
    do {
	chunk_size = size - read_size;
	if (chunk_size > BUFFER_SIZE) {
	    chunk_size = BUFFER_SIZE;
	}
	err = inflate_block(in, in->buffer, chunk_size, INFLATE_CHUNK_SIZE);
	if (err != noErr) {
	    break;
	} else {
	    DmWrite(ptr, read_size, in->buffer, chunk_size);
	    read_size += chunk_size;
	}
    } while (read_size < size);

    MemHandleUnlock(h);

    return err;
}

static int inflate_skip(struct in_stream *in, ULong size)
{
    ULong read_size, chunk_size;
    int err;

    read_size = 0;
    do {
	chunk_size = size - read_size;
	if (chunk_size > BUFFER_SIZE) {
	    chunk_size = BUFFER_SIZE;
	}
	err = inflate_block(in, in->buffer, chunk_size, INFLATE_CHUNK_SIZE);
	if (err != noErr) {
	    break;
	} else {
	    read_size += chunk_size;
	}
    } while (read_size < size);

    return err;
}


/*
 * create database
 */

static int create_database(
	UInt card_no, DatabaseHdrPtr db_header, LocalID *db_id, DmOpenRef *db)
{
    int err;

    err = DmCreateDatabase(card_no, db_header->name, db_header->creator,
	    db_header->type, db_header->attributes & dmHdrAttrResDB);
    if (err != noErr) {
	return STORAGE_ERROR;
    }

    *db_id = DmFindDatabase(card_no, db_header->name);
    if (*db_id == NULL) {
	return STORAGE_ERROR;
    }

    err = DmSetDatabaseInfo(card_no, *db_id, NULL,
	    &db_header->attributes, &db_header->version,
	    &db_header->creationDate, &db_header->modificationDate,
	    &db_header->lastBackupDate, &db_header->modificationNumber,
	    NULL, NULL, &db_header->type, &db_header->creator);
    if (err != noErr) {
	return STORAGE_ERROR;
    }

    *db = DmOpenDatabase(card_no, *db_id, dmModeReadWrite);
    if (*db == NULL) {
	return STORAGE_ERROR;
    }

    return noErr;
}


/*
 * uncompress records
 */

static int validate_record_entries(
	DatabaseHdrPtr db_header, RecordEntryPtr entries)
{
    LocalID offset, next;
    int i;

    offset = sizeof(DatabaseHdrType)
	    + sizeof(RecordEntryType) * db_header->recordList.numRecords;
    next = entries[0].localChunkID;

    if (db_header->appInfoID != NULL) {
	if (db_header->appInfoID != offset) {
	    return DATA_ERROR;
	}
	if (db_header->sortInfoID != NULL) {
	    if (db_header->sortInfoID <= offset) {
		return DATA_ERROR;
	    }
	    offset = db_header->sortInfoID;
	}
	if (next <= offset) {
	    return DATA_ERROR;
	}
    } else if (db_header->sortInfoID != NULL) {
	if (db_header->sortInfoID != offset) {
	    return DATA_ERROR;
	}
	if (next <= offset) {
	    return DATA_ERROR;
	}
    } else {
	if (next != offset) {
	    return DATA_ERROR;
	}
    }

    for (i = 0; i < db_header->recordList.numRecords; ++i) {
	offset = next;
	next = entries[i + 1].localChunkID;
	if (next < offset) {
	    return DATA_ERROR;
	}
    }

    return noErr;
}

static int inflate_records(
	struct in_stream *in, UInt card_no, DatabaseHdrPtr db_header)
{
    ULong size;
    RecordEntryPtr entries;
    LocalID db_id = NULL;
    DmOpenRef db = NULL;
    LocalID offset, next;
    VoidHand h;
    LocalID app_info;
    int i;
    UInt index, attr;
    ULong id;
    int err;

    size = sizeof(RecordEntryType) * db_header->recordList.numRecords;
    entries = MemPtrNew(size + sizeof(RecordEntryType));
    if (entries == NULL) {
	return MEMORY_ERROR;
    }

    err = inflate_block(in, entries, size + sizeof(Word), INFLATE_CHUNK_SIZE);
    entries[db_header->recordList.numRecords].localChunkID = in->uncompressed;

    if (err == noErr) {
	err = validate_record_entries(db_header, entries);
    }

    if (err == noErr) {
	err = create_database(card_no, db_header, &db_id, &db);
    }

    next = sizeof(DatabaseHdrType)
	    + sizeof(RecordEntryType) * db_header->recordList.numRecords;

    if (err == noErr && db_header->appInfoID != NULL) {
	offset = next;
	if (db_header->sortInfoID != NULL) {
	    next = db_header->sortInfoID;
	} else {
	    next = entries[0].localChunkID;
	}
	size = next - offset;
	h = DmNewHandle(db, size);
	if (h == NULL) {
	    err = STORAGE_ERROR;
	} else {
	    err = inflate_handle(in, h, size);
	    if (err == noErr) {
		app_info = MemHandleToLocalID(h);
		err = DmSetDatabaseInfo(card_no, db_id, NULL,
			NULL, NULL, NULL, NULL, NULL, NULL,
			&app_info, NULL, NULL, NULL);
		if (err != noErr) {
		    err = STORAGE_ERROR;
		}
	    }
	}
    }

    if (err == noErr && db_header->sortInfoID != NULL) {
	offset = next;
	next = entries[0].localChunkID;
	size = next - offset;
	err = inflate_skip(in, size);
    }

    if (err == noErr) {
	for (i = 0; i < db_header->recordList.numRecords; ++i) {
	    offset = next;
	    next = entries[i + 1].localChunkID;
	    size = next - offset;
	    if (size > 0) {
		index = dmMaxRecordIndex;
		h = DmNewRecord(db, &index, size);
		if (h == NULL) {
		    err = STORAGE_ERROR;
		} else {
		    err = inflate_handle(in, h, size);
		    if (err == noErr) {
			attr = entries[i].attributes;
			id = ((ULong) entries[i].uniqueID[0] << 16)
			   + ((ULong) entries[i].uniqueID[1] << 8)
			   + ((ULong) entries[i].uniqueID[2]);
			DmSetRecordInfo(db, index, &attr, &id);
		    }
		    DmReleaseRecord(db, index, true);
		}
		if (err != noErr) { break; }
	    }
	}
    }

    if (err == noErr) {
	err = DmSetDatabaseInfo(card_no, db_id, NULL,
		NULL, NULL, NULL, NULL, NULL, &db_header->modificationNumber,
		NULL, NULL, NULL, NULL);
	if (err != noErr) {
	    err = STORAGE_ERROR;
	}
    }

    if (db != NULL) {
	DmCloseDatabase(db);
    }

    if (err != noErr && db_id != NULL) {
	DmDeleteDatabase(card_no, db_id);
    }

    MemPtrFree(entries);

    return err;
}


/*
 * uncompress resources
 */

static int validate_resource_entries(
	DatabaseHdrPtr db_header, RsrcEntryPtr entries)
{
    LocalID offset, next;
    int i;

    offset = sizeof(DatabaseHdrType)
	    + sizeof(RsrcEntryType) * db_header->recordList.numRecords;
    next = entries[0].localChunkID;

    if (next != offset) {
	return DATA_ERROR;
    }

    for (i = 0; i < db_header->recordList.numRecords; ++i) {
	offset = next;
	next = entries[i + 1].localChunkID;
	if (next <= offset) {
	    return DATA_ERROR;
	}
    }

    return noErr;
}

static int inflate_resources(
	struct in_stream *in, UInt card_no, DatabaseHdrPtr db_header)
{
    ULong size;
    RsrcEntryPtr entries;
    LocalID db_id = NULL;
    DmOpenRef db = NULL;
    int i;
    VoidHand h;
    int err;

    size = sizeof(RsrcEntryType) * db_header->recordList.numRecords;
    entries = MemPtrNew(size + sizeof(RsrcEntryType));
    if (entries == NULL) {
	return MEMORY_ERROR;
    }

    err = inflate_block(in, entries, size + sizeof(Word), INFLATE_CHUNK_SIZE);
    entries[db_header->recordList.numRecords].localChunkID = in->uncompressed;

    if (err == noErr) {
	err = validate_resource_entries(db_header, entries);
    }

    if (err == noErr) {
	err = create_database(card_no, db_header, &db_id, &db);
    }

    if (err == noErr) {
	for (i = 0; i < db_header->recordList.numRecords; ++i) {
	    size = entries[i + 1].localChunkID - entries[i].localChunkID;
	    h = DmNewResource(db, entries[i].type, entries[i].id, size);
	    if (h == NULL) {
		err = STORAGE_ERROR;
	    } else {
		err = inflate_handle(in, h, size);
		DmReleaseResource(h);
	    }
	    if (err != noErr) { break; }
	}
    }

    if (err == noErr) {
	err = DmSetDatabaseInfo(card_no, db_id, NULL,
		NULL, NULL, NULL, NULL, NULL, &db_header->modificationNumber,
		NULL, NULL, NULL, NULL);
	if (err != noErr) {
	    err = STORAGE_ERROR;
	}

	/* should we do something about unique id???  */
    }

    if (db != NULL) {
	DmCloseDatabase(db);
    }

    if (err != noErr && db_id != NULL) {
	DmDeleteDatabase(card_no, db_id);
    }

    MemPtrFree(entries);

    return err;
}


/*
 * uncompress database
 */

int uncompress_db(DmOpenRef arch_db, struct stream_info *stream,
	UInt card_no, progress_callback_t callback)
{
    struct in_stream in;
    DatabaseHdrType db_header;
    int err, err2;

    err = open_stream(&in, arch_db, stream, true, callback);
    if (err != noErr) { return err; }

    err = inflate_block(&in, &db_header,
	    sizeof db_header - sizeof db_header.recordList.firstEntry,
	    INFLATE_HEADER_SIZE);

    if (err == noErr) {
	if ((db_header.attributes & dmHdrAttrResDB) == 0) {
	    err = inflate_records(&in, card_no, &db_header);
	} else {
	    err = inflate_resources(&in, card_no, &db_header);
	}

	if (err == noErr
	 && in.z.total_in != stream->size - stream->data_offset) {
	    err = DATA_ERROR;
	}
    }

    err2 = close_stream(&in);
    if (err == noErr && err2 != Z_OK) {
	err = err2;
    }

    return err;
}


/*
 * uncompress bytes
 */

int uncompress_bytes(DmOpenRef arch_db, struct stream_info *stream,
	void *ptr, ULong size, ULong skip_size, int skip_count)
{
    struct in_stream in;
    int i, err, err2;

    err = open_stream(&in, arch_db, stream, false, NULL);
    if (err != noErr) { return err; }

    for (i = 0; i < skip_count && err == noErr; ++i) {
	err = inflate_block(&in, ptr, skip_size, INFLATE_CHUNK_SIZE);
    }

    if (err == noErr) {
	err = inflate_block(&in, ptr, size, INFLATE_CHUNK_SIZE);
    }

    err2 = close_stream(&in);
    if (err == noErr && err2 != Z_OK) {
	err = err2;
    }

    return err;
}
